#include "boundparcongrad.h"
#include <string.h>
#include <math.h>
#include <time.h>

boundparcongrad::boundparcongrad(int np,int mr,long nd)
{
  nproc=np;  myrank=mr;  ndom=nd;
  ndof=0;  nbdof=0; nidof = 0;
  maxnbdof=0;  
  
  lidof= NULL;
  lbdof = NULL;
  bcn = NULL;
  nbdofdom = NULL;

 
}

boundparcongrad::~boundparcongrad()
{
}

/**
   function orders unknowns on subdomains with respect of Schur complement method
   inner variables are ordered at the beginning
   boundary variables are ordered at he end
   
   @param top - pointer to domain topology
   @param ptop - pointer to topology for parallel computation
   
   
*/
void boundparcongrad::subdomain_ordering (partop *ptop,gtopology *top,FILE *out)
{
  
  long i,j,k,ndofn,nid,ncdof;
  
  //  number of all nodes on subdomain
  nn = ptop->nn;
  //  number of internal nodes on subdomain
  nin = ptop->nin;
  //  number of boundary nodes on subdomain
  nbn = ptop->nbn;


  // ********************************
  //  ordering of internal unknowns
  // ********************************
  ndof=1;
  ncdof=0;
  for (i=0;i<nin;i++){
    //  node id (in local ordering)
    nid=ptop->lnin[i];
    //  number of DOFs at required node
    ndofn = top->give_ndofn (nid);
    
    for (j=0;j<ndofn;j++){
      k=top->give_dof (nid,j);
      if (k>0){
	top->save_dof (nid,j,ndof);
	ndof++;
      }
      else{
	ncdof++;
      }
    }
  }
  //  number of internal DOFs (unknowns)
  nidof = ndof-1;

  
  // ********************************
  //  ordering of boundary unknowns
  // ********************************
  for (i=0;i<nbn;i++){
    //  node id (in local ordering)
    nid=ptop->lnbn[i];
    //  number of DOFs at required node
    ndofn = top->give_ndofn (nid);
    
    for (j=0;j<ndofn;j++){
      k=top->give_dof (nid,j);
      if (k>0){
	top->save_dof (nid,j,ndof);
	ndof++;
      }
      else{
	ncdof++;
      }
    }
  }
  ndof--;
  

  //  contributions from elements
  for (i=0;i<top->ne;i++){
    if (top->give_cne (i)==1){
      for (j=0;j<top->give_ndofe(i);j++){
	if (top->give_dof (i,j)>ndof)  ndof=top->give_dof (i,j);
      }
    }
  }
  
  //  number of boundary DOFs (unknowns)
  nbdof = ndof-nidof;
  
  //  state of code numbers is changed
  //  code numbers are generated
  top->cnstate=1;
  
  /*
  // test print
  fprintf (out,"\n\n number of DOFs on domain is %ld\n",ndof);   
  fprintf (out,"number of constrained  DOFs on domain is %ld\n",ncdof);   
  fprintf (out,"number of internal DOFs on domain is %ld\n",nidof);   
  fprintf (out,"number of boundary DOFs on domain is %ld\n\n\n",nbdof);   
  fprintf (out,"local DOFs on domain\n");   
  
  for(i = 0; i < nn; i++){
    fprintf (out,"\n node %ld ",i+1);    
    ndofn = top->give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      fprintf (out,"    DOF  %ld  %ld",j,k);   
    }
  }
  */
  
  
}
/**
   JB 14.08.2007
  
*/


void boundparcongrad::initiate (selnodes *selnodschur,partop *ptop,gtopology *top,FILE *out)
{
  long i,j,k,l,ndofn,nsn;
  //long *aux;
  
  if (myrank==0){
    //  array containing numbers of DOFs on boundary nodes
    nbdofdom =  selnodschur -> ndofdom;
    // array containing boundary global code numbers
    bcn =  selnodschur -> cndom;
    //  number of global DOFs = total number of boundary DOFs
    ndofcp = selnodschur -> tndof;
  }
  maxnbdof =  selnodschur->maxndof;
  //  list of boundary DOFs
  lbdof = selnodschur->ldof;
  //  number of boundary degrees of freedom on subdomain
  nbdof = selnodschur->ndof;
  //  number of internal degrees of freedom on subdomain
  nidof = ndof - nbdof;
  
  //nsn = selnodschur -> nsn;
  //aux = selnodschur -> lsnl;
  
  //  list of internal DOFs
  lidof = new long[nidof];
  l = 0;
  for(i = 0; i < nn; i++){
    ndofn = top->give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      if( k <= nidof && k > 0){
	lidof[l] = k;
	l++;
      }
    }
  }
  
  /*
  fprintf (out,"\n\n number of DOFs on domain is %ld\n",ndof);   
  // test print
  fprintf (out,"\n\n number of internal DOFs on domain is %ld\n",nidof);   
  fprintf (out,"\nlist of internal DOFs on domain\n\n");   
  for(i = 0; i < nidof; i++){
    fprintf (out,"%ld %ld\n",i+1,lidof[i]);   
    }
  fprintf (out,"\n\n number of boundary DOFs on domain is %ld\n",nbdof);   
  fprintf (out,"\nlist of boundary DOFs on domain\n\n");   
  for(i = 0; i < nbdof; i++){
  fprintf (out,"%ld %ld\n",i+1,lbdof[i]);   
    } 
  */
}




/**
   function orders unknowns in coarse problem
   
   
   @param ptop - pointer to parallel topology
    
   function is not using

*/
void boundparcongrad::coarse_problem_ordering (long *domproc,gtopology *top,partop *ptop,FILE *out)
{
  long i,j,k,m,tnbn;
  long *nbnd,*ndofncn;
  long **cnbn,**nbdofnd,**cncn;
  long ***lbcn;
  MPI_Status stat;
  
  if (myrank==0){
    
    
    //  array of numbers of boundary nodes on subdomains
    nbnd = ptop->nbnd;
    //  array containing coarse node numbers of boundary nodes
    cnbn = ptop->cnbn;
    //  array containing numbers of DOFs on boundary nodes
    nbdofnd = ptop->nbdofnd;
    //  total number of boundary nodes
    tnbn = ptop->tnbn;
    //  array containing local boundary code numbers, 0 or 1
    lbcn = ptop->lbcn;
    
    if (nbnd==NULL){
      fprintf (stderr,"\n\n array containing local boundary code numbers (lbcn) is not allocated");
      fprintf (stderr,"\n in function coarse_problem_ordering (file %s, line %d)",__FILE__,__LINE__);
      fprintf (stderr,"\n function code_numbers_on_master (class partop) should be called first,");
    }
    
    //  number of DOFs at coarse nodes
    ndofncn = new long [tnbn];
    
    for (i=0;i<nproc;i++){
      for (j=0;j<nbnd[i];j++){
	k=cnbn[i][j];
	ndofncn[k]=nbdofnd[i][j];
      }
    }
    
    
    //  assignment of constraint indicators
    cncn = new long* [tnbn];
    for (i=0;i<tnbn;i++){
      cncn[i] = new long [ndofncn[i]];
    }
    
    
    //  number of boundary DOFs on subdomains
    //  it must not be deleted, it is used in further subroutines
    nbdofdom = new long [nproc];
    
    maxnbdof=0;
    for (i=0;i<nproc;i++){
      nbdofdom[i]=0;
      for (j=0;j<nbnd[i];j++){
	m=cnbn[i][j];
	for (k=0;k<nbdofnd[i][j];k++){
	  cncn[m][k]=lbcn[i][j][k];
	  if (lbcn[i][j][k]>0)
	    nbdofdom[i]++;
	}
      }
      if (maxnbdof<nbdofdom[i])
	maxnbdof=nbdofdom[i];
    }
    

    //  generation of code numbers (unknown ordering)
    ndofcp=1;
    for (i=0;i<tnbn;i++){
      for (j=0;j<ndofncn[i];j++){
	if (cncn[i][j]>0){
	  cncn[i][j]=ndofcp;
	  ndofcp++;
	}
      }
    }
    ndofcp--;
    
    
    bcn = new long* [nproc];
    for (i=0;i<nproc;i++){
      bcn[i] = new long [nbdofdom[i]];
      nbdofdom[i]=0;
    }
    
    //  assembling of lists of boundary code numbers on subdomains
    for (i=0;i<nproc;i++){
      for (j=0;j<nbnd[i];j++){
	m=cnbn[i][j];
	for (k=0;k<ndofncn[m];k++){
	  if (cncn[m][k]>0){
	    bcn[i][nbdofdom[i]]=cncn[m][k];
	    nbdofdom[i]++;
	  }
	}
      }
    }
    /*
    //  test print
    fprintf (out,"\n\n number of boundary code numbers on each subdomain\n\n\n");   
    for (i=0;i<nproc;i++){
      fprintf (out,"Domain %ld %ld\n",i+1,nbdofdom[i]);
    }
    fprintf (out,"\n\n boundary global code numbers on master\n\n\n");   
    for (i=0;i<nproc;i++){
      fprintf (out,"Domain %ld\n",i+1);
      for(j = 0; j < nbdofdom[i]; j++){
	fprintf (out,"%ld %ld\n",j+1,bcn[i][j]);
      }
    }
    */
    
    for (i=0;i<tnbn;i++){
      delete [] cncn[i];
    }
    delete [] cncn;
    
    delete [] ndofncn;
    
  }
  
  if (myrank==0){
    for (i=1;i<nproc;i++){
      MPI_Send (&maxnbdof,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  else{
    MPI_Recv (&maxnbdof,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  // test print
  fprintf (out,"\n\n maximum number of boundary DOFs is  %ld\n\n\n",maxnbdof);   
}


/*
  funkce skalarne vynasobi vnitrni DOFs vektoru u a v

  JB
*/

double boundparcongrad::v_ixu_i(double *u,double *v)
{
  long i,j;
  double sp;
  sp = 0.0;
  for(i = 0; i < nidof; i++){
    // kvuli cislovani od 0 - natace C
    j = lidof[i]-1;
    sp+=u[j]*v[j];
  }
  
  return sp;
}

/*
  funkce skalarne vynasobi vektory u a v

  JB
*/

double boundparcongrad::uxv(double *u,double *v, long n)
{
  long i;
  double sp;
  sp = 0.0;
  for(i = 0; i < n; i++){
    sp+=u[i]*v[i];
  }
  
  return sp;
  
  
}


/*
  funkce vybere do buff hranicni DOFs z vektoru v na slaves
  
  JB
*/
void boundparcongrad::select_bound(double *buff,double *v)
{
  long i,j;
  for(i = 0; i < nbdof; i++){
    // kvuli cislovani od 0 - natace C
    j = lbdof[i]-1;
    buff[i] = v[j];
  }
}

/*
  funkce prida z buff do v boundary DOFs na slaves
  
*/

void boundparcongrad::add_bound(double *buff,double *v)
{
  long i,j;
  for(i = 0; i < nbdof; i++){
    // kvuli cislovani od 0 - natace C
    j = lbdof[i]-1;
    v[j] = buff[i];
  }
}


/*
  funkce prida buff do coarse vector na mastru
  
  JB
*/

void boundparcongrad::add_master_buff (double *buff,double *cv,long nd)
{
  long i,j;
  for (i=0;i<nbdofdom[nd];i++){
    // kvuli cislovani od 0 - natace C
    j=bcn[nd][i]-1;
    cv[j]+=buff[i];
  }
  
}

/*
  funkce vybere z cv do buff odpovidajici boundary DOFs
  
*/

void boundparcongrad::select_master_buff(double *cv,double *buff,long nd)
{
  long i,j;
  for (i=0;i<nbdofdom[nd];i++){
    // kvuli cislovani od 0 - natace C
    j=bcn[nd][i]-1;
    buff[i] =  cv[j];
  }
  
}

// ********************************************************
// ********************************************************
// ********************************************************
// ********************************************************
// ********************************************************
// ********************************************************
// ********************************************************





/**
   function solves system of algebraic equations by parallel
   conjugate gradient method
   
   @param top - topology
   @param gm - %matrix of the system
   @param domproc - domain-processor correspondence
   @param lhs - array containing solution of the system
   @param rhs - array containing right hand side
   @param out - output stream
   @param iv - initial values indicator
   
   iv=0 - initial vector is zero %vector
   iv=1 - initial vector is taken from x array
   
  
*/
void boundparcongrad::solve_system (partop *ptop,gtopology *top,gmatrix *gm,
			       long *domproc,double *lhs,double *rhs,FILE *out,long iv)
{
  long i,j,k;
  long buffsize;
  double a,nom_i,denom_i,alpha,beta;
  double *p_dom, *r_dom, *d_dom,*x_dom,*buff;
  
  
  double  nom_master_i,nom_master_b,nom,denom;
  double *r_bound, *d_bound, *x_bound, *p_bound;
  double denom_master_i,denom_master_b;

  MPI_Status stat;

  // arrays defined on each subdomain
  // ndof - number of degrees of freedom
  //  auxiliary vector
  p_dom = new double [ndof];
  nullv (p_dom,ndof);
  //  residuum vector
  r_dom = new double [ndof];
  nullv (r_dom,ndof);
  //  direction vector
  d_dom = new double [ndof];
  nullv (d_dom,ndof);
  //
  x_dom = new double [ndof];
  nullv (x_dom,ndof);
  //  buffer
  buffsize=maxnbdof+2;
  //maxnbdof+1 - nom_i or denom_i
  //maxnbdof+2 - error of computation
  buff = new double [buffsize];
  
  //  initial vector
  if (iv==0){
    for (i=0;i<ndof;i++){
      lhs[i]=0.0;
    }
  }
  
  
  // vazeni prave strany
  //for(i = 0; i < ndof; i++){
  //  rhs[i]/=double(ptop->dofmultip[i]);
  //}
  

  //  matrix multiplied by initial vector
  // p_dom,j = K,j*lhs,j
  gm->gmxv (lhs,p_dom);
  
  //  initial residuum
  // r_dom,j = rhs,j - p_dom,j
  subv (rhs,p_dom,r_dom,ndof);
  
  // initial direction vector
  // d_dom,j = r_dom,j
  copyv (r_dom,d_dom,ndof);
  
  // test print
  //for(i = 0; i < ndof; i++){
  //  fprintf (out,"%ld rhs %le  p %le  r %le  d %le\n",i+1,rhs[i],p_dom[i],r_dom[i],d_dom[i]);
  //}
  
  // nom_i,j = r_i,j^T*r_i,j
  nom_i = v_ixu_i(r_dom,r_dom);

  // test print
  //fprintf (out,"nom_i %le\n",nom_i);
  
  nullv (buff,buffsize);
  // r_b,j = Lb,j*r,j
  select_bound(buff,r_dom);
  
  buff[maxnbdof] = nom_i;
  
  
  if(myrank == 0){
    
    // arrays defined only on master
    //  vector of coarse residuum
    r_bound = new double [ndofcp];
    nullv (r_bound,ndofcp);
    //  coarse direction vector
    d_bound = new double [ndofcp];
    nullv (d_bound,ndofcp);
    //  vector of unknowns
    x_bound = new double [ndofcp];
    nullv (x_bound,ndofcp);
    //  auxiliary vector
    p_bound = new double [ndofcp];
    nullv (p_bound,ndofcp);
    
    //  master contribution
    nom_master_i = 0.0;
    nom_master_i += buff[maxnbdof];
    k=domproc[0];
    // r_b=Sum_j=1^m L_b,j^T*r_b,j
    add_master_buff (buff,r_bound,k);
    
    // slave contributions
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      //nom_master_i = Sum_j=1^m nom_i,j
      nom_master_i += buff[maxnbdof];
      // r_b=Sum_j=1^m L_b,j^T*r_b,j
      add_master_buff (buff,r_bound,k);
    }
    
    //test print
    //fprintf (out,"nom_master_i %le\n",nom_master_i);
    
    // test print r_bound
    //for(i = 0; i < ndof; i++){
    //fprintf (out,"%ld r_bound %le\n",i+1,r_bound[i]);
    //}

    //nom_master_b = r_b^T*r_b
    nom_master_b = uxv(r_bound,r_bound,ndofcp);
    
    // test print
    //fprintf (out,"nom_master_b %le\n",nom_master_b);
    
    // nom = nom_master_i+nom_master_b
    nom =  nom_master_i+nom_master_b;
    
    //test print nom
    //fprintf (out,"nom %le\n",nom);
    
    // d_b = r_b
    copyv (r_bound,d_bound,ndofcp);
    
    //test print d_bound
    //for(i = 0; i < ndofcp; i++){
    //fprintf (out,"%ld r_bound %lf d_bound %lf\n",i+1,r_bound[i],d_bound[i]); 
    //}
    
    //d_b,j=L_b,s*d_b
    // slaves
    for (i=1;i<nproc;i++){
      k=domproc[i];
      nullv (buff,maxnbdof);
      select_master_buff(d_bound,buff,k);
      MPI_Send (buff,buffsize,MPI_DOUBLE,i,myrank,MPI_COMM_WORLD); 
    }
    
    // master
    k=domproc[0];
    nullv (buff,buffsize);
    select_master_buff(d_bound,buff,k);
    
  }
  else{
    
    MPI_Send (buff,buffsize,MPI_DOUBLE,0,myrank,MPI_COMM_WORLD);  
    MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  // d_dom,j = d_i,j, d_b,j
  add_bound(buff,d_dom);
  
  //test print
  // for(i = 0; i < ndof; i++){
  // fprintf (out,"%ld d_dom %lf\n",i+1,d_dom[i]); 
  // }
  
  // iteration loop
  for(i = 0; i < ni; i++){
    
    //  p_dom,j = Kj*d_dom,j
    gm->gmxv (d_dom,p_dom);
    
    //  denom_i,j = s_i,j^T*d_i,j
    denom_i =  v_ixu_i(d_dom,p_dom);
    
    // test print
    //fprintf (out,"denom_i %lf\n",denom_i);
    
    nullv (buff,maxnbdof);
    // p_b,j = Lb,j*r,j
    select_bound(buff,p_dom);
    
    buff[maxnbdof] = denom_i;
    
    
    if(myrank == 0){
       
      nullv (p_bound,ndofcp);
       
      //  master contribution
      k=domproc[0];
      denom_master_i = 0.0;
      denom_master_i += buff[maxnbdof];
      // p_b=Sum_j=1^m L_b,j^T*p_b,j
      add_master_buff (buff,p_bound,k);
      
      // slave contributions
      for (j=1;j<nproc;j++){
	MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	k=domproc[stat.MPI_TAG];
	//denom_master_i = Sum_j=1^m nom_i,j
	denom_master_i += buff[maxnbdof];
	// p_b=Sum_j=1^m L_b,j^T*p_b,j
	add_master_buff (buff,p_bound,k);
      }
      
      //test print p_bound
      //for(j = 0; j < ndofcp; j++){
      //fprintf (out,"%ld p_bound %lf\n",j+1,p_bound[j]); 
      //}
      
      //nom_master_b = r_b^T*r_b
      denom_master_b = uxv(d_bound,p_bound,ndofcp);
     
      // nom = nom_master_i+nom_master_b
      denom =  denom_master_i+denom_master_b;
      
      //test print
      //fprintf (out,"denom %le\n",denom);
      
      // alpha = nom/denom;
      alpha = nom/denom;
      denom = nom;
      
      //test print
      //fprintf (out,"alpha %le\n",alpha);
      
      // new approximation of x_b and r_b
      // r_b^k+1 = r_b^k-alpha*p_b
      // x_b^k+1 = x_b^k+alpha*d_b
      for (j=0;j<ndofcp;j++){
	x_bound[j]+=alpha*d_bound[j];
	r_bound[j]-=alpha*p_bound[j];
      }
      
      //nom_master_b = r_b^T*r_b
      nom_master_b = uxv(r_bound,r_bound,ndofcp);
      
      for (j=1;j<nproc;j++){
	MPI_Send (&alpha,1,MPI_DOUBLE,j,myrank,MPI_COMM_WORLD); 
      } 
      
    }
    else{
      MPI_Send (buff,buffsize,MPI_DOUBLE,0,myrank,MPI_COMM_WORLD);   
      MPI_Recv (&alpha,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat); 
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    
    // new approximation of x_dom and r_dom - internal DOFs
    for (j=0;j<nidof;j++){
      // kvuli cislovani v C
      k = lidof[j]-1;
      x_dom[k]+=alpha*d_dom[k];
      r_dom[k]-=alpha*p_dom[k];
    }

    //test print
    //for(j = 0; j < ndofcp; j++){
    //  fprintf (out,"%ld r_bound %lf\n",j+1,r_bound[j]); 
    //}
    //for(j = 0; j < ndofcp; j++){
    //  fprintf (out,"%ld x_bound %lf\n",j+1,x_bound[j]); 
    // }
     
    // nom_i,j^k+1 = (r_i,j^k+1)^T*r_i,j^k+1
    nom_i = v_ixu_i(r_dom,r_dom);
    a = nom_i;

    if(myrank == 0){
      
      nom_master_i = 0.0;
      // master contributions
      nom_master_i = a;
      // slave contributions
      for (j=1;j<nproc;j++){
	MPI_Recv (&a,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	k=domproc[stat.MPI_TAG];
	nom_master_i += a;
      }
      
      // nom = nom_master_i+nom_master_b
      nom =  nom_master_i+nom_master_b; 
      
      //test print
      //fprintf (out,"nom k+1 %le\n",nom);
      
      if (nom<err || i == ni-1 ){
	buff[maxnbdof+1]=10000.0;
      }
      
      fprintf (stdout,"\n iteration number   %ld    nom %le",i,nom);
      
      beta = nom/denom;
      
      //test print
      //fprintf (out,"beta %le\n",beta);
      
      //  new aproximation of vector d_b
      for (j=0;j<ndofcp;j++){
	d_bound[j]=beta*d_bound[j]+r_bound[j];
      }
      
      //test print
      //for(j = 0; j < ndofcp; j++){
      //fprintf (out,"%ld d_bound %lf\n",j+1,d_bound[j]); 
      //}
      
      for (j=1;j<nproc;j++){
	k=domproc[j];
	nullv (buff,maxnbdof);
	//d_b,j=L_b,s*d_b
	select_master_buff(d_bound,buff,k);
	
	buff[maxnbdof]=beta;
	
	MPI_Send (buff,buffsize,MPI_DOUBLE,j,myrank,MPI_COMM_WORLD); 
      }
      
      k=domproc[0];
      nullv (buff,maxnbdof);
      //d_b,j=L_b,s*d_b 
      select_master_buff(d_bound,buff,k);
      
      buff[maxnbdof] = beta;
      
    }
    else{
      MPI_Send (&a,1,MPI_DOUBLE,0,myrank,MPI_COMM_WORLD); 
      MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    
    // d_dom,j = d_i,j, d_b,j
    add_bound(buff,d_dom);
    
    beta = buff[maxnbdof];
    
    //fprintf (out,"beta_dom %le\n",beta);
    
    //  new aproximation of vector d_dom,j
    for (j=0;j<nidof;j++){
      // kvuli cislovani v C
      k = lidof[j]-1;
      d_dom[k]=beta*d_dom[k]+r_dom[k];
    }
    
    if (buff[maxnbdof+1]==10000.0){
      if (myrank==0){
	// slaves
	for (j=1;j<nproc;j++){
	  k=domproc[j];
	  nullv (buff,maxnbdof);
	  select_master_buff(x_bound,buff,k);
	  MPI_Send (buff,buffsize,MPI_DOUBLE,j,myrank,MPI_COMM_WORLD); 
	}
	// master
	k=domproc[0];
	nullv (buff,buffsize-1);
	select_master_buff(d_bound,buff,k);
	
	// deleting
	delete []r_bound;
	delete []d_bound;
	delete []x_bound;
	delete []p_bound;
      
      }
      else{
	MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      }
      // x 
      add_bound(buff,x_dom);
      //
      //for (j=0;j<ndof;j++){
	//fprintf (out,"%ld %le\n",j+1,x_dom[j]); 
      //}
      
      // results of computation
      copyv (x_dom,lhs,ndof);
      
      // deleting
      delete []r_dom;
      delete []d_dom;
      delete []x_dom;
      delete []p_dom;
      
      break;
    }
    
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  
}
